from ..date_utils import add_months, period_resolution
from ..triangle import Triangle

__all__ = ["shift_origin"]


def shift_origin(triangle: Triangle, origin_match_triangle: Triangle):
    """Shift the origin of a triangle to match the origin of another triangle.

    Args:
        triangle: The triangle to shift.
        origin_match_triangle: The triangle to match the origin of.

    Returns:
        Triangle: The shifted triangle.
    """
    tri_resolution = period_resolution(triangle)
    match_resolution = period_resolution(origin_match_triangle)
    if tri_resolution != match_resolution:
        raise ValueError(
            f"Cannot shift triangle with resolution {tri_resolution} to match resolution {match_resolution}"
        )
    if tri_resolution not in [3, 12]:
        raise ValueError("Only annual and quarterly triangles are supported")
    tri_origin = triangle.periods[0][0].month % tri_resolution
    match_origin = origin_match_triangle.periods[0][0].month % match_resolution
    month_shift = match_origin - tri_origin
    shifted_tri = triangle.replace(
        period_start=lambda cell: add_months(cell.period_start, month_shift),
        period_end=lambda cell: add_months(cell.period_end, month_shift),
        evaluation_date=lambda cell: add_months(cell.evaluation_date, month_shift),
    )
    return shifted_tri



from .__about__ import __version__
import os

from .base import Cell, CumulativeCell, IncrementalCell, Metadata

from .errors import *
from .io import *
from .matrix import (
    DisaggregatedValue,
    Matrix,
    MatrixIndex,
    MissingValue,
    PredictedValue,
    RichMatrix,
)

from .utils import *
from .factory import Triangle
from .triangle import TriangleSlice
from .date_utils import *

local_dir = os.path.dirname(os.path.abspath(__file__))
meyers_tri = Triangle.from_binary(os.path.join(local_dir, "meyers.trib"))



import datetime
from typing import Callable

import toolz as tlz

from ..base import Cell, CumulativeCell, IncrementalCell, Metadata
from ..date_utils import (
    add_months,
    dev_lag_months,
    eval_date_resolution,
    period_resolution,
    standardize_resolution,
)
from ..errors import TriangleError
from ..triangle import Triangle
from .basis import to_cumulative, to_incremental

__all__ = [
    "make_right_triangle",
    "make_right_diagonal",
    "make_pred_triangle",
    "make_pred_triangle_complement",
    "make_pred_triangle_with_init",
]


def make_right_triangle(
    triangle: Triangle[Cell], dev_lags=None, dev_lag_unit="month"
) -> Triangle:
    """Create the lower-right complement of a triangle."""

    cum_tri = to_cumulative(triangle) if triangle.is_incremental else triangle
    new_cells = tlz.mapcat(
        tlz.partial(_make_right_triangle_slice, dev_lags, dev_lag_unit),
        cum_tri.slices.values(),
    )
    right_tri = Triangle(list(new_cells))
    if triangle.is_incremental:
        right_tri = _fix_prev_evaluation_date(triangle, right_tri.to_incremental())
    return right_tri


def _make_right_triangle_slice(dev_lags, unit, slice_):
    if dev_lags is None:
        dev_lags = {cell.dev_lag(unit) for cell in slice_}

    return [
        CumulativeCell(
            period_start=cell.period_start,
            period_end=cell.period_end,
            evaluation_date=_add_dev_lag(cell, dev_lag, unit),
            metadata=cell.metadata,
            values={},
        )
        for dev_lag in dev_lags
        for cell in slice_.right_edge
        if dev_lag > cell.dev_lag(unit)
    ]


def _add_dev_lag(cell, dev_lag, unit):
    _unit = unit.lower()
    if "month" in _unit:
        return add_months(cell.period_end, dev_lag)
    elif "day" in _unit:
        return cell.period_end + datetime.timedelta(dev_lag)
    elif _unit == "timedelta":
        return cell.period_end + datetime.timedelta(dev_lag)
    else:
        raise ValueError(
            f"Unrecognized unit '{unit}', must be 'month', 'day' or 'timedelta'."
        )


def make_right_diagonal(triangle, evaluation_dates, include_historic=False) -> Triangle:
    """Create a new `Triangle` with one or more diagonals to the right of existing data."""

    cum_tri = triangle.to_cumulative() if triangle.is_incremental else triangle
    new_cells = tlz.mapcat(
        tlz.partial(_make_right_diagonal_slice, evaluation_dates, include_historic),
        cum_tri.slices.values(),
    )
    right_tri = Triangle(list(new_cells))
    if triangle.is_incremental:
        right_tri = _fix_prev_evaluation_date(triangle, right_tri.to_incremental())
    return right_tri


def _make_right_diagonal_slice(
    evaluation_dates, include_historic, slice_
) -> list[Cell]:
    max_eval = max(cell.evaluation_date for cell in slice_)
    if not include_historic:
        evaluation_dates = [
            eval_date for eval_date in evaluation_dates if eval_date > max_eval
        ]

    return [
        CumulativeCell(
            period_start=cell.period_start,
            period_end=cell.period_end,
            evaluation_date=eval_date,
            metadata=cell.metadata,
            values={},
        )
        for cell in slice_.right_edge.cells
        for eval_date in evaluation_dates
        if eval_date >= cell.period_start
    ]


def _fix_prev_evaluation_date(
    triangle: Triangle[IncrementalCell],
    right_tri: Triangle[IncrementalCell],
) -> Triangle[IncrementalCell]:
    edge_cells = []
    for slc in right_tri.slices.values():
        for period, row in slc.period_rows:
            edge_cells.append(row[0])

    fixed_edge_cells = []
    for cell in triangle.right_edge:
        fixed_edge_cells += [
            ob.replace(prev_evaluation_date=cell.evaluation_date)
            for ob in edge_cells
            if ob.period == cell.period and ob.metadata == cell.metadata
        ]

    return Triangle(
        [cell for cell in right_tri if cell not in edge_cells] + fixed_edge_cells
    )


def make_pred_triangle(
    metadata_sets: list[Metadata],
    min_period: datetime.date,
    max_period: datetime.date,
    exp_resolution: tuple[int, str],
    eval_resolution: tuple[int, str],
    exp_origin: datetime.date | None = None,
    eval_origin: datetime.date | None = None,
    min_dev_lag: tuple[int, str] | None = (0, "months"),
    max_dev_lag: tuple[int, str] | None = None,
    min_eval: datetime.date | None = None,
    max_eval: datetime.date | None = None,
    statics_fn: Callable[[Cell], dict[str, float]] | None = None,
    is_incremental: bool | None = False,
) -> Triangle:
    if exp_origin is None:
        exp_origin = min_period + datetime.timedelta(days=-1)
    if eval_origin is None:
        eval_origin = exp_origin

    # Assemble all of the experience periods in the data set
    exp_resolution_months = standardize_resolution(exp_resolution)[0]
    current_end = exp_origin
    periods = []
    while current_end < max_period:
        next_begin = current_end + datetime.timedelta(days=1)
        next_end = add_months(current_end, exp_resolution_months)
        if next_end <= max_period:
            periods.append((next_begin, next_end))
        current_end = next_end

    # Assemble all of the evaluation dates in the data set
    eval_resolution_months = standardize_resolution(eval_resolution)[0]
    if max_eval is None and max_dev_lag is None:
        raise Exception("One of `max_eval` or `max_dev_lag` must be specified")
    if max_eval is None:
        max_dev_lag_months = standardize_resolution(max_dev_lag)[0]
        max_eval = add_months(periods[-1][1], max_dev_lag_months)
    if max_dev_lag is None:
        max_dev_lag_months = dev_lag_months(periods[0][1], max_eval)
    else:
        max_dev_lag_months = standardize_resolution(max_dev_lag)[0]
    min_dev_lag_months = standardize_resolution(min_dev_lag)[0]
    if min_eval is None:
        min_eval = add_months(periods[0][1], min_dev_lag_months)

    current_eval = min_eval
    eval_dates = []
    while current_eval <= max_eval:
        eval_dates.append(current_eval)
        current_eval = add_months(current_eval, eval_resolution_months)

    def _cell_with_statics(period_start, period_end, evaluation_date, metadata):
        ob = CumulativeCell(
            period_start=period_start,
            period_end=period_end,
            evaluation_date=evaluation_date,
            metadata=metadata,
            values={},
        )

        try:
            return ob.replace(values=statics_fn(ob))
        except (KeyError, IndexError):
            return None

    pred_cells = [
        _cell_with_statics(period_start, period_end, evaluation_date, metadata)
        for period_start, period_end in periods
        for evaluation_date in eval_dates
        for metadata in metadata_sets
        if min_dev_lag_months
        <= dev_lag_months(period_end, evaluation_date)
        <= max_dev_lag_months
    ]
    pred_triangle = Triangle([cell for cell in pred_cells if cell is not None])

    if is_incremental:
        pred_triangle = pred_triangle.to_incremental()
    return pred_triangle


def make_pred_triangle_complement(
    init_triangle: Triangle[Cell],
    static_fields: list[str] | None = None,
    max_dev_lag: int | None = None,
    eval_date_resolution_override: int | None = None,
) -> Triangle:
    evaluation_date_res = (
        eval_date_resolution(init_triangle)
        if eval_date_resolution_override is None
        else eval_date_resolution_override
    )
    if evaluation_date_res is None:
        raise TriangleError(
            "Must provide `init_triangle` with different evaluation dates, or the evaluation date "
            "resolution of the triangle manually via. eval_date_resolution_override."
        )

    static_fields = (
        static_fields if static_fields else ["earned_premium", "earned_exposure"]
    )

    def statics_fn(empty_cell: Cell) -> dict[str, float]:
        edge_cell = [
            cell
            for cell in init_triangle.right_edge.cells
            if cell.details == empty_cell.details
            and cell.period_start == empty_cell.period_start
        ][0]
        return {k: v for k, v in edge_cell.values.items() if k in static_fields}

    if max_dev_lag is None:
        max_dev_lag = init_triangle.dev_lags()[-1]

    raw_cells = make_pred_triangle(
        metadata_sets=init_triangle.metadata,
        min_period=init_triangle.periods[0][0],
        max_period=init_triangle.periods[-1][-1],
        exp_resolution=(period_resolution(init_triangle), "months"),
        eval_resolution=(evaluation_date_res, "months"),
        exp_origin=init_triangle.periods[0][0] - datetime.timedelta(days=1),
        eval_origin=init_triangle.evaluation_dates[0] - datetime.timedelta(days=1),
        min_dev_lag=(init_triangle.dev_lags()[0], "months"),
        max_dev_lag=(max_dev_lag, "months"),
        min_eval=init_triangle.evaluation_dates[0],
        statics_fn=statics_fn,
        is_incremental=init_triangle.is_incremental,
    )

    init_cells = {}
    for cell in init_triangle:
        frozen_details = frozenset(cell.details.items())
        if frozen_details not in init_cells:
            init_cells[frozen_details] = []
        init_cells[frozen_details].append((cell.period_start, cell.evaluation_date))
    right_edge_start = {
        slice_period_id: cell[-1].evaluation_date
        for slice_period_id, cell in init_triangle.slice_period_rows
    }

    return Triangle(
        [
            cell
            for cell in raw_cells
            if (
                (cell.period_start, cell.evaluation_date)
                not in init_cells[frozenset(cell.details.items())]
            )
            and cell.evaluation_date
            > right_edge_start.get((cell.metadata, cell.period), datetime.date.min)
        ]
    )


def make_pred_triangle_with_init(
    init_triangle: Triangle[Cell],
    pred_triangle: Triangle[Cell] | None = None,
    max_dev_lag: tuple[int, str] | None = None,
    eval_resolution: tuple[int, str] | None = None,
    max_eval_date: datetime.date | None = None,
) -> Triangle:
    # Return pred_triangle if provided and no conflicts
    if pred_triangle is not None:
        if max_eval_date is not None:
            raise TriangleError(
                "`pred_triangle` and `max_eval_date` cannot both be specified"
            )
        if max_dev_lag is not None:
            raise TriangleError(
                "`pred_triangle` and `max_dev_lag` cannot both be specified"
            )
        if eval_resolution is not None:
            raise TriangleError(
                "`pred_triangle` and `eval_resolution` cannot both be specified"
            )
        if type(init_triangle.cells[0]) != type(pred_triangle.cells[0]):  # noqa: E721
            raise TriangleError(
                "`init_triangle` and `pred_triangle` have different cell types"
            )
        return pred_triangle

    # Fill in default values for other args if not supplied
    if max_dev_lag is None:
        raise TriangleError(
            "`max_dev_lag` must be supplied if `pred_triangle` is missing"
        )
    if eval_resolution is None:
        raise TriangleError(
            "`eval_resolution` must be supplied if `pred_triangle` is missing"
        )
    if max_eval_date is None:
        max_eval_date = datetime.date.max

    dev_lags = _get_all_lag_months(eval_resolution, max_dev_lag)
    pred_tri = init_triangle.make_right_triangle(dev_lags)
    return pred_tri.filter(lambda cell: cell.evaluation_date <= max_eval_date)


def _get_all_lag_months(eval_resolution, max_dev_lag):
    standardized_resolution = standardize_resolution(eval_resolution)
    standardized_max_lag = standardize_resolution(max_dev_lag)
    if standardized_resolution[1] != "month":
        raise TriangleError(
            "Cannot construct a `pred_triangle` with non-monthly resolution"
        )
    if standardized_max_lag[1] != "month":
        raise TriangleError(
            "Cannot construct a `pred_triangle` with non-monthly max dev lag"
        )
    lag_months = []
    lag = 0
    while lag <= standardized_max_lag[0]:
        lag_months.append(lag)
        lag += standardized_resolution[0]
    return lag_months



import numpy as np
import scipy.stats as stat

from ..date_utils import eval_date_resolution
from ..triangle import Triangle


def bootstrap(
    triangle: Triangle,
    n: int,
    seed: int,
    field: str | list[str] | None = None,
) -> list[Triangle]:
    """Sample bootstrap replicates from a triangle.

    Traditional bootstrapping would sample
    triangle cells with replacement, but this is
    inappropriate because cells are not iid
    realizations. This function implements
    two different types of bootstrapping
    that preserve structure between cells:

        * If a triangle has multiple development
          lags and experience periods, the 'ata'
          method is implemented, which resamples
          volume-weighted age-to-age factors
          between experience periods.
        * If the triangle has a single experience
          period, diagonal, or development
          lag, then maximum entropy bootstrapping
          is used. Maximum entropy bootstrapping (Vinod, 2006)
          is a bootstrapping method for time series
          data that does not require stationarity,
          detrending, or block bootstrapping.

    The function returns a multi-slice triangle
    with metadata details key-value pair ('bootstrap', i),
    where i is the i-th replicate.

    Arguments:
        triangle: The triangle to re-sample.
        n: The number of boostrap replicates to sample.
        seed: The RNG seed.
        field: The field, or list of fields, to resample
            in the triangle. Defaults to `None`, which
            selects all fields.

    Returns:
        A list of (potentially multi-slice) triangles, one for
        each bootstrap replicate.
    """

    if n <= 0:
        raise ValueError("Argument `n` to `bootstrap` must be a positive integer.")
    boots = (
        _bootstrap_slice(triangle_slice, n, seed, field)
        for _, triangle_slice in triangle.slices.items()
    )
    zipped = zip(*boots)
    return [sum(boot) for boot in zipped]


def _bootstrap_slice(
    triangle: Triangle,
    n: int,
    seed: int,
    field: str | list[str] | None = None,
) -> Triangle:
    """Bootstrap a single triangle slice.

    Returns:
        A multi-slice triangle, one slice per bootstrap replicate.
    """
    rng = np.random.default_rng(seed)
    fields = (
        [field]
        if isinstance(field, str)
        else field
        if isinstance(field, list)
        else triangle.fields
    )
    if (
        len(triangle.dev_lags()) > 1
        and len(triangle.periods) > 1
        and len(triangle.evaluation_dates) > 1
    ):
        method = "atas"
    else:
        method = "maximum_entropy"
    if method == "atas":
        atas, volume_weights = _empirical_atas(triangle, fields)
        S = {
            lag: {field: len(v) for field, v in ata.items()}
            for lag, ata in atas.items()
        }
        weights = {
            i: {
                lag: {field: volume_weights[lag][field] for field, s in count.items()}
                for lag, count in S.items()
            }
            for i in range(n)
        }
        resampled_atas = {
            i: {
                lag: {
                    field: atas[lag][field][
                        rng.choice(range(len(p)), size=len(p), p=p, replace=True)
                    ]
                    for field, p in w.items()
                }
                for lag, w in lag_atas.items()
            }
            for i, lag_atas in weights.items()
        }
        replicates = (
            _develop_triangle_by_atas(triangle, resampled_atas[i]).derive_metadata(
                details=lambda cell: {**cell.metadata.details, **{"bootstrap": i}}
            )
            for i in range(n)
        )
    elif method == "maximum_entropy":
        replicates = []
        for i in range(n):
            meboot = {
                field: maximum_entropy_ensemble(
                    [cell[field] for cell in triangle],
                    rng.uniform(size=len(triangle)),
                    L=(0, max(cell[field] for cell in triangle)),
                )
                for field in fields
            }
            cells = []
            for cell_idx, cell in enumerate(triangle):
                cells.append(
                    cell.derive_fields(
                        **{field: values[cell_idx] for field, values in meboot.items()}
                    )
                )
            replicates += [
                Triangle(cells).derive_metadata(
                    details=lambda cell: {**cell.metadata.details, "bootstrap": i}
                )
            ]
    return replicates


def _develop_triangle_by_atas(triangle, resampled_atas) -> Triangle:
    cells = []
    period_lookup = {period: i for i, period in enumerate(triangle.periods)}
    initial_values = {
        period: min(triangle.filter(lambda ob: ob.period == period).dev_lags())
        for period in period_lookup
    }
    for cell in triangle:
        if cell.dev_lag() == initial_values[cell.period]:
            cells.append(cell)
            values = cells[-1].values
        else:
            period_idx = period_lookup[cell.period]
            values = {
                **cell.values,
                **{
                    field: (
                        v * resampled_atas[cell.dev_lag()][field][period_idx]
                        if cell.values.get(field)
                        else None
                    )
                    for field, v in values.items()
                },
            }
            cells.append(
                cell.replace(
                    values=values,
                )
            )
    return Triangle(cells)


def _empirical_atas(triangle: Triangle, fields: list[str]):
    atas = {}
    zipped_lags = zip(triangle.dev_lags()[1:], triangle.dev_lags()[:-1])
    resolution = eval_date_resolution(triangle)

    def _safe_ata_division(x, y):
        safe_x = 1 if x is None or not x else x
        safe_y = 1 if y is None or not y else y
        return safe_x / safe_y

    for lag, prev_lag in zipped_lags:
        clipped_tri = triangle.clip(min_dev=prev_lag, max_dev=lag)
        lag_cells = [
            (next_cell, prev_cell)
            for next_cell, prev_cell in zip(clipped_tri[1:], clipped_tri[:-1])
            if next_cell.period == prev_cell.period
        ]
        atas[lag] = {
            field: np.array(
                [
                    _safe_ata_division(
                        next_cell.values.get(field), prev_cell.values.get(field)
                    )
                    for next_cell, prev_cell in lag_cells
                ]
            )
            for field in fields
        }
    volume_weight = {
        lag: {
            field: _normalize(
                np.array(
                    [
                        cell[field]
                        for cell in triangle.filter(
                            lambda cell: cell.dev_lag() == lag - resolution
                        )
                    ]
                )[: len(vals)]
            )
            for period in triangle.periods
            for field, vals in field_vals.items()
        }
        for lag, field_vals in atas.items()
    }
    return atas, volume_weight


def _normalize(x: np.array):
    if sum(x) == 0:
        return np.repeat(1 / x.size, x.size)
    return x / x.sum()


def maximum_entropy_ensemble(
    x: list[float], U: list[float], L: tuple[float, float] | None = None
) -> np.ndarray:
    """Applies the maximum entropy bootstrap algorithm of Vinod (2006).

    See the following links for Python and R implementations
    that this function is heavily based on. However, note that the
    Python implementations are wrong in that they don't appear to
    sort the final quantiles:
        * https://gist.github.com/TomKealy/975a7f639dec4febad5964950023a6c0
        * https://github.com/kirajcg/pymeboot/blob/master/pymeboot/meboot.py
        * https://github.com/cran/meboot/blob/master/R/meboot.R

    - Vinod, H.D. 2006. Maximum entropy ensembles for time series inference
        in economics. Journal of Asian Economics. 17 (6), 955-978.
    """
    LIMITS = 0.1
    if len(x) == 1:
        return x
    if all(x[0] == i for i in x[1:]):
        return x
    if any(i is None for i in x):
        raise ValueError("Cannot bootstrap missing values using maximum entropy.")
    if any(0 > u > 1 for u in U):
        raise ValueError(
            "U should be a list of draws from the Uniform(0, 1) distribution."
        )

    sorted_x, indices = map(np.array, zip(*sorted([v, i] for i, v in enumerate(x))))
    if L is None:
        trimmed_mean = stat.trim_mean(abs(np.diff(x)), LIMITS)
    z_t = [
        L[0] if L else sorted_x[0] - trimmed_mean,
        *(sorted_x[:-1] + sorted_x[1:]) / 2,
        L[1] if L else sorted_x[-1] + trimmed_mean,
    ]
    desired_means = [
        0.75 * sorted_x[0] + 0.25 * sorted_x[1],
        *(0.25 * sorted_x[:-2] + 0.5 * sorted_x[1:-1] + 0.25 * sorted_x[2:]),
        0.75 * sorted_x[-1] + 0.25 * sorted_x[-2],
    ]
    xr = np.linspace(0, 1, len(x) + 1)
    U = sorted(U)
    inds = np.searchsorted(xr, U, side="right") - 1
    interpolated = [desired_means[i] - (z_t[i] + z_t[i + 1]) / 2 for i in inds]
    y0, y1 = zip(
        *[
            (z_t[i] + interpolated[j], z_t[i + 1] + interpolated[j])
            for i, j in zip(inds, range(len(x)))
        ]
    )
    quantiles = [
        (y0[j] + ((U[j] - xr[i]) * (y1[j] - y0[j])) / (xr[i + 1] - xr[i]))
        for i, j in zip(inds, range(len(x)))
    ]
    replicate = [q for _, q in sorted(zip(indices, sorted(quantiles)))]
    return replicate



import numpy as np

from ..base import CumulativeCell, IncrementalCell
from ..date_utils import add_months, id_to_month, is_triangle_monthly, resolution_delta
from ..matrix import (
    DisaggregatedPredictedValue,
    DisaggregatedValue,
    Matrix,
    MatrixIndex,
    MissingValue,
    PredictedValue,
)
from ..triangle import Triangle

__all__ = [
    "triangle_to_rich_matrix",
    "rich_matrix_to_triangle",
]


def triangle_to_rich_matrix(
    tri: Triangle,
    eval_resolution: int | None = None,
    fields: list[str] | None = None,
) -> Matrix:
    """Convert a triangle to a rich matrix format.

    The rich matrix format is similar to the matrix format, but it somewhat more flexible. It
    can handle missing data, data that needs to be disaggregated, and predicted cells.
    """
    # Ensure that the triangle meets basic data standards
    if not is_triangle_monthly(tri):
        raise ValueError(
            "Triangle cannot be converted to matrices.  The `period_start` of each triangle "
            "cell must be on the first day of the month.  Also, the `period_end` and "
            "`evaluation_date`s of each triangle cell must be on the last day of the month.  "
            "One or more of these conditions is not satisfied for %s" % tri
        )

    # Build the index
    index = MatrixIndex.from_triangle(
        tri, eval_resolution=eval_resolution, fields=fields
    )
    next_disagg_id = 0

    # Figure out maximum dimensions
    if fields is None:
        if not tri.fields:
            raise Exception("Must specify fields if triangle has none")
        fields = tri.fields

    _, _, max_period, max_dev = index.resolve_indices(
        tri.metadata[0], fields[0], tri.periods[-1][0], tri.dev_lags()[-1]
    )
    data_dims = (len(tri.metadata), len(fields), max_period + 1, max_dev + 1)
    covered_dims = (len(tri.metadata), max_period + 1, max_dev + 1)

    # Fill in the data
    data = np.full(data_dims, None, dtype=object)
    is_covered = np.full(covered_dims, False)
    for cell in tri:
        for field, raw_value in cell.values.items():
            if field in fields:
                # Grab indices for the cell
                meta_ndx, field_ndx, exp_start_ndx, dev_ndx = index.resolve_indices(
                    cell.metadata, field, cell.period_start, cell.dev_lag()
                )
                _, _, exp_end_ndx, _ = index.resolve_indices(
                    cell.metadata, field, cell.period_end, cell.dev_lag()
                )

                # Mark the cells as covered
                for i, exp_ndx in enumerate(range(exp_end_ndx, exp_start_ndx - 1, -1)):
                    is_covered[meta_ndx, exp_ndx, dev_ndx + i] = True
                # Fill in the matrix with values
                if isinstance(raw_value, np.ndarray):
                    value = (
                        float(raw_value)
                        if raw_value.size == 1
                        else PredictedValue(raw_value)
                    )
                else:
                    value = raw_value
                if exp_start_ndx == exp_end_ndx:
                    data[meta_ndx, field_ndx, exp_start_ndx, dev_ndx] = value
                else:
                    disagg_value = (
                        DisaggregatedPredictedValue(
                            id=next_disagg_id, value=value.value
                        )
                        if isinstance(value, PredictedValue)
                        else DisaggregatedValue(id=next_disagg_id, value=value)
                    )
                    next_disagg_id += 1
                    for i, exp_ndx in enumerate(
                        range(exp_end_ndx, exp_start_ndx - 1, -1)
                    ):
                        data[meta_ndx, field_ndx, exp_ndx, dev_ndx + i] = disagg_value

    # Fill in missing values that are covered by other cells
    next_missing_id = 0
    for meta_ndx in range(data.shape[0]):
        for exp_ndx in range(data.shape[2]):
            for dev_ndx in range(data.shape[3]):
                if is_covered[meta_ndx, exp_ndx, dev_ndx]:
                    for field_ndx in range(data.shape[1]):
                        if data[meta_ndx, field_ndx, exp_ndx, dev_ndx] is None:
                            data[meta_ndx, field_ndx, exp_ndx, dev_ndx] = MissingValue(
                                next_missing_id
                            )
                            next_missing_id += 1

    return Matrix(data=data, index=index, incremental=tri.is_incremental)


def rich_matrix_to_triangle(mat: Matrix) -> Triangle:
    period_starts = [
        id_to_month(mat.index._exp_origin + i * mat.index._exp_resolution)
        for i in range(mat.data.shape[2])
    ]
    period_ends = [
        id_to_month(
            mat.index._exp_origin + (i + 1) * mat.index._exp_resolution - 1, False
        )
        for i in range(mat.data.shape[2])
    ]
    min_resolution = min(mat.index._exp_resolution, mat.index._dev_resolution)
    dev_lags = [
        float(mat.index._dev_origin + i * min_resolution)
        for i in range(mat.data.shape[3])
    ]

    cells = []

    for i, meta in enumerate(mat.index._slices):
        for j, (period_start, period_end) in enumerate(zip(period_starts, period_ends)):
            for k, dev_lag in enumerate(dev_lags):
                raw_values = mat.data[i, :, j, k]
                values = {
                    name: (val.value if isinstance(val, PredictedValue) else val)
                    for name, val in zip(mat.index._fields, raw_values)
                    if val is not None
                    and not isinstance(
                        val,
                        (MissingValue, DisaggregatedValue, DisaggregatedPredictedValue),
                    )
                }
                if values:
                    if not mat.incremental:
                        new_cell = CumulativeCell(
                            period_start=period_start,
                            period_end=period_end,
                            evaluation_date=add_months(period_end, dev_lag),
                            metadata=meta,
                            values=values,
                        )
                    else:
                        prev_eval = (
                            resolution_delta(period_start, (-1, "days"))
                            if not k
                            else add_months(period_end, dev_lags[k - 1])
                        )
                        new_cell = IncrementalCell(
                            period_start=period_start,
                            period_end=period_end,
                            evaluation_date=add_months(period_end, dev_lag),
                            prev_evaluation_date=prev_eval,
                            metadata=meta,
                            values=values,
                        )
                    cells.append(new_cell)

    return Triangle(cells)



from collections import defaultdict
from warnings import warn

from ..base import Cell
from ..triangle import Triangle
from .join import join

__all__ = [
    "merge",
    "period_merge",
    "loose_period_merge",
    "coalesce",
]


def merge(
    tri1: Triangle,
    tri2: Triangle,
    join_type: str = "full",
    on: list[str] | None = None,
) -> Triangle:
    """Merge cells of two triangles, if coordinates overlap then prioritize right cell values.

    All cells from tri1 and tri2 are preserved by this merge. If two cells overlap by index,
    (period_start, period_end, evaluation_date, metadata), then values from tri2 take
    precedence.

    Merge is a higher-level operation that returns a Triangle of merged cells. For a
    lower-level implementation that returns a list of pairs of matched cells, see `join`.

    `merge` is different from `coalesce` in that it combines the union of fields in
    cells at the same coordinates, where `coalesce` only keeps the fields and values from
    the cell with precedence.

    Args:
        tri1: The lower-precedence, left triangle to merge.
        tri2: The higher-precedence, right triangle to merge.
        join_type: The type of relational join to perform. For a list of all join types
            and their descriptions, see the documentation for `join`.
        on: which metadata attributes and detail keys to join on when merging cell values.
            Default behavior is to consider all metadata attributes when merging cells.
    Returns:
        A triangle with the merged cells from the two triangles.
    """
    cell_pairs = join(tri1, tri2, join_type, on)
    return Triangle([_merge_cell_pair(*cell_pair) for cell_pair in cell_pairs])


def _merge_cell_pair(cell1: Cell | None, cell2: Cell | None) -> Cell:
    if cell1 is None:
        return cell2
    elif cell2 is None:
        return cell1
    else:
        # We can safely assume that cell1 and cell2 cannot both be None
        return cell1.replace(values={**cell1.values, **cell2.values})


def period_merge(
    tri1: Triangle[Cell],
    tri2: Triangle[Cell],
    suffix: str | None = None,
) -> Triangle[Cell]:
    """Add period-level values of tri2 to tri1 cells. tri2 must have a single cell per period index.

    All cells from tri1 are preserved by this merge. If there is a matching cell in tri2 by index
    (period_start, period_end, metadata), then values from tri2 take precedence. There must be
    only one cell in tri2 per index.

    Args:
        tri1: First triangle in the merge.
        tri2: Second triangle in the merge.
        suffix: If present, then every field in tri2 will have the suffix appended to the name
            of the field after the merge to avoid name conflicts. If None, then no suffix
            will be appended and the values in tri2 will overwrite the values in tri1.
    """
    # Make sure that both triangles have the same cell type
    if min(len(tri1), len(tri2)) > 0 and type(tri1.cells[0]) != type(tri2.cells[0]):  # noqa: E721
        raise ValueError("Triangles must have the same cell type")

    tri1_cells = defaultdict(list)
    for cell in tri1:
        tri1_cells[(cell.period_start, cell.period_end, cell.metadata)] += [cell]

    tri2_cells = defaultdict(list)
    for cell in tri2:
        tri2_cells[(cell.period_start, cell.period_end, cell.metadata)] += [cell]

    output_cells = []
    for idx, cells in tri1_cells.items():
        right_cell = tri2_cells[idx]
        if len(right_cell) > 1:
            raise (ValueError(f"multiple cells at tri2 index {idx}"))
        elif len(right_cell) == 0:
            output_cells += cells
        else:
            output_cells += [
                _overwrite_values(cell, right_cell[0], suffix=suffix) for cell in cells
            ]
    return Triangle(output_cells)


def _overwrite_values(cell1: Cell, cell2: Cell, suffix: str | None = None) -> Cell:
    """Replace cell1 values with cell2 values for common fields."""
    if suffix:
        replace_map = {f"{k}{suffix}": v for k, v in cell2.values.items()}
    else:
        replace_map = cell2.values

    return cell1.replace(values={**cell1.values, **replace_map})


def loose_period_merge(
    tri1: Triangle[Cell],
    tri2: Triangle[Cell],
    suffix: str | None = None,
) -> Triangle[Cell]:
    """Add period-level values of tri2 to tri1 cells. tri2 must have a single cell per period index.

    All cells from tri1 are preserved by this merge. If there is a matching cell in tri2 by index
    (period_start, period_end, metadata), then values from tri2 take precedence. There must be
    only one cell in tri2 per index.

    Note that this is different from `period_merge` in that it will allow partial matches on
    metadata details. For examples, if tri1 has a cell with metadata details `{"a": 1, "b": 2}`
    and tri2 has a cell with metadata details `{"a": 1}`, then the cell from tri2 will be merged
    with the cell from tri1. This is useful for merging triangles with different levels of
    granularity, particularly in BFing. Tri1 must be more granular (or as granular) as tri2.

    Args:
        tri1: First triangle in the merge.
        tri2: Second triangle in the merge.
        suffix: If present, then every field in tri2 will have the suffix appended to the name
            of the field after the merge to avoid name conflicts. If None, then no suffix
            will be appended and the values in tri2 will overwrite the values in tri1.
    """
    # Make sure that both triangles have the same cell type
    if min(len(tri1), len(tri2)) > 0 and not isinstance(
        tri1.cells[0], type(tri2.cells[0])
    ):
        raise ValueError("Triangles must have the same cell type")
    if tri1.cells[0].metadata.details.keys() < tri2.cells[0].metadata.details.keys():
        raise ValueError("tri1 must have at least as many metadata details as tri2")

    common_details = tri1.metadata[0].details.keys() & tri2.metadata[0].details.keys()
    tri1_cells = defaultdict(list)
    for cell in tri1:
        metadata_proxy = cell.derive_metadata(
            details=lambda meta: {k: meta.details[k] for k in common_details}
        )
        tri1_cells[(cell.period_start, cell.period_end, metadata_proxy.metadata)] += [
            cell
        ]

    tri2_cells = defaultdict(list)
    for cell in tri2:
        tri2_cells[(cell.period_start, cell.period_end, cell.metadata)] += [cell]

    output_cells = []
    for idx, cells in tri1_cells.items():
        right_cell = tri2_cells[idx]
        if len(right_cell) > 1:
            raise (ValueError(f"multiple cells at tri2 index {idx}"))
        elif len(right_cell) == 0:
            output_cells += cells
        else:
            output_cells += [
                _overwrite_values(cell, right_cell[0], suffix=suffix) for cell in cells
            ]
    return Triangle(output_cells)


def coalesce(triangles: list[Triangle]) -> Triangle:
    """Coalesce triangle cells from a list of Triangles such that the resulting
    triangle is a union of the cells from each triangle. Matching cells in the
    earlier triangles in the triangle list take precedence over cells from later
    triangles.

    `coalesce` is different from `merge` in that it only keeps the fields and values from
    the cell with precedence, wheras `merge` will return the union of fields across the
    matching cells.

    Args:
        triangles: A list of Triangle whose cells will be coalesced.

    Returns:
        A Triangle with Cells at all coordinates of the triangles in the triangle
        list.
    """
    if not isinstance(triangles, list):
        raise ValueError("coalesce must be provided a list of triangles")
    grouped_cells = defaultdict(list)
    for triangle in triangles:
        for cell in triangle:
            grouped_cells[(cell.metadata, cell.period, cell.evaluation_date)].append(
                cell
            )

    if not any(len(matching_cells) > 1 for matching_cells in grouped_cells.values()):
        warn(
            """
        No overlapping cells in the list of triangles provided, check for differences
        in metadata
        """
        )
    return Triangle([cell_options[0] for cell_options in grouped_cells.values()])



import numpy as np

from ..triangle import Cell, Triangle


def thin(triangle: Triangle, num_samples: int, seed: int = None) -> Triangle:
    """'Thin out' a triangle by replacing all array-typed values with a
    random subset of values. For example, given a triangle `pred_tri` with
    10,000 samples per value, `thin(pred_tri, 500)` will return
    a triangle with 500 samples per value.

    The thinning process takes the same random subset of samples from each
    value and preserves the relative order of the subsamples, so that
    correlations between values (both within and across cells) are preserved.

    Arguments:
        triangle: The triangle to thin.
        num_samples: The number of samples of each array-valued value to return.
        seed: The seed of the random number generator.
    Returns:
        A copy of the triangle with `num_samples` random samples.
    """

    if triangle.num_samples < num_samples:
        raise ValueError(
            f"Cannot return {num_samples} samples, "
            f"the source triangle only has {num_samples} samples"
        )
    elif triangle.num_samples == num_samples:
        return triangle
    else:
        rng = np.random.default_rng(seed)
        ndxs = rng.choice(triangle.num_samples, num_samples, False)
        return Triangle([_thin_cell(cell, ndxs) for cell in triangle])


def _thin_cell(cell: Cell, ndxs: np.ndarray) -> Cell:
    new_values = {
        k: v[ndxs] if isinstance(v, np.ndarray) and len(v) > 1 else v
        for k, v in cell.values.items()
    }
    return cell.replace(values=new_values)



import numpy as np

from ..base import CumulativeCell, IncrementalCell
from ..date_utils import add_months, id_to_month, is_triangle_monthly, resolution_delta
from ..matrix import Matrix, MatrixIndex
from ..triangle import Triangle

__all__ = [
    "triangle_to_matrix",
    "matrix_to_triangle",
]


def triangle_to_matrix(
    tri: Triangle,
    eval_resolution: int | None = None,
    fields: list[str] | None = None,
) -> Matrix:
    """Convert a triangle to a matrix format.

    The matrix format is well-suited for clean data with consistent experience period lengths
    and evaluation date deltas. If a triangle is not regular or if it has Predictions, it
    cannot be converted to a matrix.
    """
    # Ensure that the triangle meets basic data standards
    if not is_triangle_monthly(tri):
        raise Exception("Only monthly Triangles can be converted to matrices")
    if not tri.is_semi_regular():
        raise Exception("Only semi-regular Triangles can be converted to matrices")

    if fields is None:
        if not tri.fields:
            raise Exception("Must specify fields if triangle has none")
        fields = tri.fields

    # Build the index
    index = MatrixIndex.from_triangle(
        tri, eval_resolution=eval_resolution, fields=fields
    )

    # Figure out maximum dimensions
    _, _, max_period, max_dev = index.resolve_indices(
        tri.metadata[0], fields[0], tri.periods[-1][0], tri.dev_lags()[-1]
    )
    data_dims = (len(tri.metadata), len(fields), max_period + 1, max_dev + 1)

    # Fill in the data
    data = np.full(data_dims, np.nan)
    for cell in tri:
        for field, value in cell.values.items():
            if field in fields:
                data[
                    index.resolve_indices(
                        cell.metadata, field, cell.period_start, cell.dev_lag()
                    )
                ] = float(value)

    return Matrix(data=data, index=index, incremental=tri.is_incremental)


def matrix_to_triangle(mat: Matrix) -> Triangle:
    period_starts = [
        id_to_month(mat.index._exp_origin + i * mat.index._exp_resolution)
        for i in range(mat.data.shape[2])
    ]
    period_ends = [
        id_to_month(
            mat.index._exp_origin + (i + 1) * mat.index._exp_resolution - 1, False
        )
        for i in range(mat.data.shape[2])
    ]
    dev_lags = [
        float(mat.index._dev_origin + i * mat.index._dev_resolution)
        for i in range(mat.data.shape[3])
    ]

    cells = []

    for i, meta in enumerate(mat.index._slices):
        for j, (period_start, period_end) in enumerate(zip(period_starts, period_ends)):
            for k, dev_lag in enumerate(dev_lags):
                raw_values = mat.data[i, :, j, k]
                values = {
                    name: val
                    for name, val in zip(mat.index._fields, raw_values)
                    if not np.isnan(val)
                }
                if values:
                    if not mat.incremental:
                        new_cell = CumulativeCell(
                            period_start=period_start,
                            period_end=period_end,
                            evaluation_date=add_months(period_end, dev_lag),
                            metadata=meta,
                            values=values,
                        )
                    else:
                        prev_eval = (
                            resolution_delta(period_start, (-1, "days"))
                            if not k
                            else add_months(period_end, dev_lags[k - 1])
                        )
                        new_cell = IncrementalCell(
                            period_start=period_start,
                            period_end=period_end,
                            evaluation_date=add_months(period_end, dev_lag),
                            prev_evaluation_date=prev_eval,
                            metadata=meta,
                            values=values,
                        )
                    cells.append(new_cell)

    return Triangle(cells)



from .array import (
    array_data_frame_to_triangle,
    array_triangle_builder,
    statics_data_frame_to_triangle,
    triangle_to_array_data_frame,
    triangle_to_right_edge_data_frame,
)
from .binary_input import binary_to_triangle
from .binary_output import triangle_to_binary
from .chain_ladder import chain_ladder_to_triangle, triangle_to_chain_ladder
from .data_frame_input import (
    long_csv_to_triangle,
    long_data_frame_to_triangle,
    wide_csv_to_triangle,
    wide_data_frame_to_triangle,
)
from .data_frame_output import (
    triangle_to_long_csv,
    triangle_to_long_data_frame,
    triangle_to_wide_csv,
    triangle_to_wide_data_frame,
)
from .json import (
    dict_to_triangle,
    json_string_to_triangle,
    json_to_triangle,
    triangle_json_load,
    triangle_json_loads,
    triangle_to_dict,
    triangle_to_json,
)
from .matrix import *
from .rich_matrix import *



import datetime
from warnings import warn

import numpy as np

from ..base import Metadata
from ..date_utils import eval_date_resolution, month_to_id, period_resolution
from ..triangle import Triangle

__all__ = [
    "MatrixIndex",
]


class MatrixIndex(object):
    """Utility class to handle natural indexing of Matrix-style Bermuda data."""

    def __init__(
        self,
        slices: list[Metadata],
        fields: list[str],
        exp_origin: int,
        dev_origin: int,
        exp_resolution: int,
        dev_resolution: int,
    ):
        self._slices = slices
        self._slice_lookup = {meta: ndx for ndx, meta in enumerate(self._slices)}
        self._fields = fields
        self._field_lookup = {field: ndx for ndx, field in enumerate(self._fields)}
        self._exp_origin = exp_origin
        self._exp_resolution = exp_resolution
        self._dev_origin = dev_origin
        self._dev_resolution = dev_resolution

    slices = property(lambda self: self._slices)
    slice_lookup = property(lambda self: self._slice_lookup)
    fields = property(lambda self: self._fields)
    field_lookup = property(lambda self: self._field_lookup)
    exp_origin = property(lambda self: self._exp_origin)
    exp_resolution = property(lambda self: self._exp_resolution)
    dev_origin = property(lambda self: self._dev_origin)
    dev_resolution = property(lambda self: self._dev_resolution)

    @classmethod
    def from_triangle(
        cls,
        tri: Triangle,
        eval_resolution: int | None = None,
        fields: list[str] | None = None,
    ) -> "MatrixIndex":
        # Get experience indices
        exp_origin = month_to_id(min([start for start, _ in tri.periods]))
        exp_resolution = period_resolution(tri)
        # Get dev lag indices
        int_dev_lags = [int(lag) for lag in tri.dev_lags()]
        dev_origin = min(int_dev_lags)
        dev_resolution = eval_date_resolution(tri)
        if not fields:
            fields = tri.fields
            if not fields:
                raise Exception("Must provide fields if Triangle has none")

        if not eval_resolution:
            if not dev_resolution:
                raise Exception(
                    "Must supply eval_resolution for single dev lag Triangle"
                )
            eval_resolution = dev_resolution
        else:
            if dev_resolution and eval_resolution > dev_resolution:
                warn(
                    f"Lowering resolution from {dev_resolution} months to {eval_resolution} months"
                )

        return cls(
            slices=tri.metadata,
            fields=fields,
            exp_origin=exp_origin,
            exp_resolution=exp_resolution,
            dev_origin=dev_origin,
            dev_resolution=eval_resolution,
        )

    def subset(self, shape, *args) -> "MatrixIndex":
        if len(args) != 4:
            raise Exception("Subsetting expects exactly 4 arguments")
        slice_ndxs, field_ndxs, exp_ndxs, dev_ndxs = args
        slice_ndxs = [slice_ndxs] if isinstance(slice_ndxs, int) else slice_ndxs
        field_ndxs = [field_ndxs] if isinstance(field_ndxs, int) else field_ndxs

        if isinstance(exp_ndxs, slice):
            first_exp = 0 if exp_ndxs.start is None else exp_ndxs.start
        else:
            first_exp = exp_ndxs
        if isinstance(dev_ndxs, slice):
            first_dev = 0 if dev_ndxs.start is None else dev_ndxs.start
        else:
            first_dev = dev_ndxs

        _, _, exp_size, dev_size = shape
        if first_exp < 0:
            first_exp = exp_size - first_exp
        if first_dev < 0:
            first_dev = dev_size - first_dev

        return MatrixIndex(
            slices=[self._slices[ndx] for ndx in slice_ndxs],
            fields=[self._fields[ndx] for ndx in field_ndxs],
            exp_origin=self._exp_origin + self._exp_resolution * first_exp,
            exp_resolution=self._exp_resolution,
            dev_origin=self._dev_origin + self._dev_resolution * first_dev,
            dev_resolution=self._dev_resolution,
        )

    def resolve_indices(self, *args):
        if len(args) != 4:
            raise Exception("Matrix indexing expects exactly 4 arguments")
        slice_, field, exp, dev = args
        return (
            self._resolve_index(slice_, self._resolve_slice_ndx),
            self._resolve_index(field, self._resolve_field_ndx),
            self._resolve_index(exp, self._resolve_exp_ndx, False),
            self._resolve_index(dev, self._resolve_dev_ndx, False),
        )

    @staticmethod
    def _resolve_index(arg, ndx_fn, allow_iterable=True):
        if isinstance(arg, (list, np.ndarray)):
            if allow_iterable:
                return [ndx_fn(elem) for elem in arg]
            else:
                raise Exception(
                    "Indexing by an iterable on this dimension is not allowed"
                )
        elif isinstance(arg, slice):
            return slice(ndx_fn(arg.start), ndx_fn(arg.stop), arg.step)
        else:
            return ndx_fn(arg)

    def _resolve_slice_ndx(self, arg):
        if arg is None:
            return None
        if isinstance(arg, int):
            if arg < 0:
                raise Exception(f"Slice index {arg} is out of range")
            return arg
        elif isinstance(arg, Metadata):
            return self._slice_lookup[arg]
        else:
            raise Exception(f"Unknown slice indexing type {arg.__class__.__name__}")

    def _resolve_field_ndx(self, arg):
        if arg is None:
            return None
        elif isinstance(arg, int):
            if arg < 0:
                raise Exception(f"Field index {arg} is out of range")
            return arg
        elif isinstance(arg, str):
            return self._field_lookup[arg]
        else:
            raise Exception(f"Unknown field indexing type {arg.__class__.__name__}")

    def _resolve_exp_ndx(self, arg):
        if arg is None:
            return None
        elif isinstance(arg, int):
            if arg < 0:
                raise Exception(f"Experience index {arg} is out of range")
            return arg
        elif isinstance(arg, datetime.date):
            arg_id = month_to_id(arg)
            arg_ndx = (arg_id - self._exp_origin) // self._exp_resolution
            if arg_ndx < 0:
                raise Exception(f"Experience date {arg} is out of range")
            return arg_ndx
        else:
            raise Exception(
                f"Unknown experience period indexing type {arg.__class__.__name__}"
            )

    def _resolve_dev_ndx(self, arg):
        if arg is None:
            return None
        elif isinstance(arg, float):
            # Triangles may have unequal experience period and development lag
            # resolutions, so we need to choose the smallest resolution
            # to create the resolved index.
            ndx_resolution = min(self._dev_resolution, self._exp_resolution)
            arg_ndx = int((arg - self._dev_origin) / ndx_resolution)
            if arg_ndx < 0:
                raise Exception(f"Development lag {arg} is out of range")
            return arg_ndx
        elif isinstance(arg, int):
            if arg < 0:
                raise Exception(f"Development index {arg} is out of range")
            return arg
        else:
            raise Exception(f"Unknown dev lag indexing type {arg.__class__.__name__}")



import datetime
import re
from warnings import warn

import pandas as pd

from ..base import CumulativeCell, Metadata
from ..date_utils import add_months, calculate_dev_lag
from ..triangle import Triangle
from ..utils import merge


def array_triangle_builder(
    dfs: list[pd.DataFrame],
    fields: list[str],
    **kwargs,
) -> Triangle:
    """
    Build a Triangle object from a list of DataFrames in a triangular array format.

    Args:
        dfs: List of DataFrames in a triangular array format.
        fields: List of field names.
        **kwargs: Additional arguments to pass to array_data_frame_to_triangle.
    """
    if not len(dfs) == len(fields):
        raise ValueError("The number of DataFrames and fields must match.")
    triangle = array_data_frame_to_triangle(dfs[0], fields[0], **kwargs)
    for df, field in zip(dfs[1:], fields[1:]):
        triangle = merge(triangle, array_data_frame_to_triangle(df, field, **kwargs))
    return triangle


def array_data_frame_to_triangle(
    df: pd.DataFrame,
    field: str,
    period_resolution: int | None = None,
    eval_resolution: int | None = None,
    dev_lag_from_period_end: bool = True,
    metadata: Metadata | None = None,
) -> Triangle:
    """
    Convert a DataFrame in a triangular array format to a Triangle object. The first
    column is assumed to contain the accident or policy period. This should be a date,
    or a string that can be parsed as a date (e.g. "2020", or "2020Q1"). All subsequent
    columns are assumed to contain values for the development periods. If integer development
    lags cannot be inferred from the column names an eval_resolution should be provided.

    Args:
        df: DataFrame in a triangular array format.
        field: Field name.
        period_resolution: Period resolution in months. If None, this is inferred from
            the first consecutive period dates.
        eval_resolution: Evaluation resolution in months. If None, this is inferred from
            the column names. If None and column names are missing, it's assumed to be
            the same as period_resolution.
        dev_lag_from_period_end: Is the development lag calculated from the period_end
            rather than the period start? The bermuda convention is to calculate dev
            lags as the number of months from the end of the period (so the first natural
            evaluation for any period is lag 0). If False, the dev lag is calculated from
            the period start.
        metadata: Metadata associated with the triangle.
    """

    df = df.rename(columns={df.columns[0]: "period"})
    df["period"] = df["period"].apply(parse_date)
    if period_resolution is None:
        if df.shape[0] < 2:
            raise ValueError(
                "At least two periods are required to infer period resolution."
            )
        period_resolution = int(
            calculate_dev_lag(df["period"].iloc[0], df["period"].iloc[1])
        )
    if eval_resolution is None:
        try:
            [int(x) for x in df.columns[1:]]
        except ValueError:
            warn(
                (
                    "Could not infer evaluation resolution from column names - "
                    f"defaulting to period resolution {period_resolution}"
                )
            )
            eval_resolution = period_resolution

    if metadata is None:
        metadata = Metadata()

    cells = []
    for _, row in df.iterrows():
        period_start = row["period"]
        period_end = add_months(row["period"], period_resolution) - datetime.timedelta(
            days=1
        )
        current_lag = 0
        for dev_lag in df.columns[1:]:
            if eval_resolution is None:
                current_lag = int(dev_lag)
            value = row[dev_lag]
            if dev_lag_from_period_end or eval_resolution is not None:
                # Typical case where we're using the bermuda dev lag convention
                # or just inferring the eval resolution from the period resolution
                evaluation_date = add_months(period_end, current_lag)
            else:
                # If specified by the user we can handle dev lags using the
                # period_start as 0
                evaluation_date = add_months(
                    period_start, current_lag
                ) - datetime.timedelta(days=1)
            if not pd.isna(value):
                cells.append(
                    CumulativeCell(
                        period_start=period_start,
                        period_end=period_end,
                        evaluation_date=evaluation_date,
                        values={field: row[dev_lag]},
                        metadata=metadata,
                    )
                )
            if eval_resolution is not None:
                current_lag += eval_resolution
    return Triangle(cells)


def statics_data_frame_to_triangle(
    df: pd.DataFrame,
    evaluation_date: datetime.date | None = None,
    period_resolution: int | None = None,
    metadata: Metadata | None = None,
):
    """
    Convert a DataFrame in a "statics" format to a Triangle object. The first
    column is assumed to contain the accident or policy period. This should be a date,
    or a string that can be parsed as a date (e.g. "2020", or "2020Q1"). All subsequent
    columns are assumed to contain values for a single development period that are typically
    assumed to be static over development periods (e.g. earned premium, earned exposure).
    In order to convert these values into a triangle an evaluation_date must be provided. If
    the resulting triangle can often be added to an existing loss triangle using the add_statics
    function, in which case the evaluation_date is just a placeholder (though this must be greater
    than the latest period). If none is provided we will use the last period end as the evaluation
    date. The period resolution can be inferred from the first two consecutive periods, otherwise
    it must be provided as an integer. Metadata can be provided to add additional information to
    the cells. Field names will be inferred from the column names.

    Args:
        df: DataFrame to convert.
        evaluation_date: Evaluation date for the static values. This is typically a placeholder
            date that is greater than the latest period in the triangle.
        period_resolution: Period resolution in months. If None, this is inferred from
            the first consecutive period dates.
        metadata: Metadata associated with the triangle.
    """
    df = df.rename(columns={df.columns[0]: "period"})
    df["period"] = df["period"].apply(parse_date)
    if period_resolution is None:
        if df.shape[0] < 2:
            raise ValueError(
                "At least two periods are required to infer period resolution."
            )
        period_resolution = (df["period"].iloc[1] - df["period"].iloc[0]).days // 30
    if evaluation_date is None:
        evaluation_date = add_months(
            df["period"].max(), period_resolution
        ) - datetime.timedelta(days=1)
    if metadata is None:
        metadata = Metadata()

    cells = []
    for _, row in df.iterrows():
        period_start = row["period"]
        period_end = add_months(row["period"], period_resolution) - datetime.timedelta(
            days=1
        )
        cells.append(
            CumulativeCell(
                period_start=period_start,
                period_end=period_end,
                evaluation_date=evaluation_date,
                values=row[1:].to_dict(),
                metadata=metadata,
            )
        )
    return Triangle(cells)


def parse_date(value):
    """Convert various date formats into datetime.date objects."""
    if pd.isna(value):
        return None

    value = str(value).strip()

    try:
        return pd.to_datetime(value).date()
    except (ValueError, TypeError):
        pass

    if re.fullmatch(r"\d{4}", value):
        return datetime.date(int(value), 1, 1)

    match = re.fullmatch(r"(\d{4})Q([1-4])", value)
    if match:
        year, quarter = int(match.group(1)), int(match.group(2))
        month = (quarter - 1) * 3 + 1
        return datetime.date(year, month, 1)

    match = re.fullmatch(r"(\d{4})H([1-2])", value)
    if match:
        year, half = int(match.group(1)), int(match.group(2))
        month = 1 if half == 1 else 7
        return datetime.date(year, month, 1)
    raise ValueError(f"Could not parse date: {value}")


def triangle_to_array_data_frame(triangle: Triangle, field: str):
    """
    Convert a Triangle object to a DataFrame in a triangular array format.

    Args:
        triangle: A single-sliced Triangle object.
        field: Field name.
    """
    filtered_tri = triangle.filter(lambda cell: field in cell.values)
    if len(triangle.slices) > 1:
        raise ValueError("Triangle must be single-sliced.")
    if triangle.is_incremental:
        raise ValueError("Triangle must be cumulative.")
    rows = []
    for period, cells in filtered_tri.period_rows:
        row = {"period": period[0]}
        for cell in cells:
            row[str(int(cell.dev_lag()))] = cell[field]
        rows.append(row)
    df = pd.DataFrame(rows)
    return df


def triangle_to_right_edge_data_frame(triangle: Triangle) -> pd.DataFrame:
    """
    Convert a Triangle object to a DataFrame displaying values
    on the right edge of the triangle. The resulting dataframe will
    show the 'period', 'evaluation_date', and cell values for the
    right edge of the triangle.

    Args:
        triangle: A single-sliced Triangle object.
    """
    if len(triangle.slices) > 1:
        raise ValueError("Triangle must be single-sliced.")
    if triangle.is_incremental:
        raise ValueError("Triangle must be cumulative.")
    rows = [
        {
            "period": cell.period_start,
            "evaluation_date": cell.evaluation_date,
            **cell.values,
        }
        for cell in triangle.right_edge
    ]
    return pd.DataFrame(rows)
